/*

  pyhack.c - Python XScreensaver Hack Module and GLHack Class

  Copyright (C) 2004 Alastair Tse <acnt2@cam.ac.uk>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <sys/time.h>
#include <sys/param.h>
#include <getopt.h>

#include "pyhack.h"
#include "hack.h"


char *pyhack_name = "pyhack";
int signalled = 0;

void __sig_handler(int sig) {
	 signalled = 1;
}


static PyObject * GLHack_new(PyTypeObject *type, 
							 PyObject *args,
							 PyObject *kwds) {
 
  GLHackObject *self;
  self = (GLHackObject *)type->tp_alloc(type, 0);
  if (self != NULL) {
	self->usingRootWindow = 0;
	self->frameTime = 0;
	self->be_nice = 0;
	self->FIXME_name = "GLHack";
  }
  return (PyObject *)self;
}

static void GLHack_dealloc(GLHackObject *self) {

  /* cleanup X display */
  if (self->display) {
	if (self->window) {
	  if (!((self->usingRootWindow) || (self->existingWindow)))
		XDestroyWindow (self->display, self->window);
	}
	XCloseDisplay(self->display);
  }

}

GLHACK_NOARGS_FUNC(GLHack_draw) {
  /* override me! */
  //printf("calling GLHack_draw!\n");
  return  Py_None;
}

GLHACK_ARG_FUNC(GLHack_reshape) {
  /* override me! */
  //printf("calling GLHack_reshape\n");
  return Py_None;
}

GLHACK_ARG_FUNC(GLHack_handle_opts) {
  int argc, i, ok = 0;
  char **argv = NULL;
  PyObject *arguments = NULL, *rr = NULL;

  static struct option long_options[] = {
	GLHACK_OPTIONS_LONG
  };
  int c;

  // extract argc, argv
  ok = PyArg_ParseTuple(args, "O", &arguments);

  if (ok && PyList_Check(arguments)) {
	argc = PyList_Size(arguments);
	argv = calloc(sizeof(char *), argc);
	for (i = 0; i < PyList_Size(arguments); i++) {
	  PyObject *a = PyList_GetItem(arguments, i);
	  if (PyString_Check(a))
		argv[i] = PyString_AsString(a);
	}
  }
  else {
	PyErr_SetString(PyExc_TypeError, "Failed to parse args");
	return NULL;
  }  

  while (1) {
	c = getopt_long(argc, argv, GLHACK_OPTIONS_SHORT, long_options, NULL);
	if (c == -1)
	  break;
	switch (c) {
	  GLHACK_OPTIONS_CASES;
	}
  }

  return Py_None;
}

GLHACK_NOARGS_FUNC(GLHack_run) {

  int bFPS = 0;
  int fps = 0;
  struct timeval then, now, fps_time;
  XEvent event;
  Atom XA_WM_PROTOCOLS = XInternAtom (self->display, \
									  "WM_PROTOCOLS", False);
  Atom XA_WM_DELETE_WINDOW = XInternAtom (self->display, \
										  "WM_DELETE_WINDOW", False);

  if (!self->usingRootWindow) {
	XSetWMProtocols (self->display, self->window, &XA_WM_DELETE_WINDOW, 1);
  }

  hack_clear_buffers(self);

  gettimeofday (&now, NULL);
  while (!signalled) {
	PyObject *r = PyObject_CallMethod((PyObject *)self, "draw", NULL);
	if (r == NULL) {
	  return NULL;
	}
	Py_XDECREF(r);

	glXSwapBuffers (self->display, self->window);

	if (bFPS) {
	  if (fps != -1)
		fps++;
	  
	  gettimeofday (&now, NULL);
	  
	  if (now.tv_sec > fps_time.tv_sec) {
		if (fps != -1) {
		  printf ("%d fps\n", fps);
		}
		
		fps = 0;
		fps_time.tv_sec = now.tv_sec;
		fps_time.tv_usec = now.tv_usec;
	  }
	}

	while (XPending (self->display)) {
	  KeySym keysym;
	  char c = 0;
	  
	  XNextEvent (self->display, &event);
	  switch (event.type) {
	  case ConfigureNotify:
		if ((int)self->windowWidth != event.xconfigure.width \
			|| (int)self->windowHeight != event.xconfigure.height) {
		  PyObject *rr = NULL;
		  self->windowWidth = event.xconfigure.width;
		  self->windowHeight = event.xconfigure.height;
		  
		  hack_clear_buffers(self);
		  rr = PyObject_CallMethod((PyObject *)self, \
								   "reshape", "(ii)", \
								   self->windowWidth, self->windowHeight);
		  Py_XDECREF(rr);
		}
		
		break;
	  case KeyPress:
		XLookupString (&event.xkey, &c, 1, &keysym, 0);
		
		if (c == 'f') {
		  bFPS = !bFPS;
		  
		  if (bFPS) {
			fps = -1;
			gettimeofday (&fps_time, NULL);
		  }
		}
		
		if (c == 'q' || c == 'Q' || c == 3 || c == 27)
		  return Py_None;
		
		break;
	  case ClientMessage:
		if (event.xclient.message_type == XA_WM_PROTOCOLS) {
		  if (event.xclient.data.l[0] == XA_WM_DELETE_WINDOW) {
			return Py_None;
		  }
		}
		break;
	  case DestroyNotify:
		return Py_None;
	  }
	}


	if (self->frameTime) {
	  int frameTimeSoFar = 0;
	  then = now;
	  gettimeofday (&now, NULL);
	  frameTimeSoFar = (now.tv_sec - then.tv_sec) * 1000000 + now.tv_usec - then.tv_usec;

	  while (frameTimeSoFar < self->frameTime) {
		if (self->be_nice) {
		  struct timeval tv;
		  Py_BEGIN_ALLOW_THREADS
		  tv.tv_sec = 0;
		  tv.tv_usec = self->frameTime - frameTimeSoFar;
		  select (0, 0, 0, 0, &tv);
		  Py_END_ALLOW_THREADS
		}
		gettimeofday (&now, NULL);
		frameTimeSoFar = (now.tv_sec - then.tv_sec) * 1000000 + now.tv_usec - then.tv_usec;
	  }
	} else if (self->be_nice) {
	  struct timeval tv;
	  Py_BEGIN_ALLOW_THREADS
	  tv.tv_sec = 0;
	  tv.tv_usec = 1000;
	  select (0, 0, 0, 0, &tv);
	  Py_END_ALLOW_THREADS
	}

	

  }

  return Py_None;
}




static int GLHack_init(GLHackObject *self, PyObject *args) {

  int argc, i,j, ok = 0;
  char **argv = NULL;
  PyObject *arguments = NULL, *rr = NULL;
  struct sigaction sa;
  char *display_name = NULL;
  PyObject *newargv = NULL;
  
  // extract argc, argv
  ok = PyArg_ParseTuple(args, "O", &arguments);

  if (ok && PyList_Check(arguments)) {
	argc = PyList_Size(arguments);
	argv = calloc(sizeof(char *), argc);
	for (i = 0; i < PyList_Size(arguments); i++) {
	  PyObject *a = PyList_GetItem(arguments, i);
	  if (PyString_Check(a))
		argv[i] = PyString_AsString(a);
	}
  }
  else {
	PyErr_SetString(PyExc_TypeError, "Failed to parse args");
	return -1;
  }

  Py_XDECREF(arguments);

  self->existingWindow = 0;
  self->display = NULL;
  self->window = 0;

  /* extract -window-id from the argc/argv */
  for (i = 0; i < argc; i++) {
	if (!strcmp (argv[i], "-window-id")) {
	  if ((argv[i + 1][0] == '0') \
		  && ((argv[i + 1][1] == 'x') \
			  || (argv[i + 1][1] == 'X'))) {
		
		self->existingWindow = strtol ((char *)(argv[i + 1] + 2), \
									   (char **)NULL, 16);
	  } 
	  else {
		self->existingWindow = strtol ((char *)(argv[i + 1]), \
									   (char **)NULL, 10);
	  }

	  for (j = i + 2; j < argc; j++) {
		argv[j - 2] = argv[j];
	  }
	  
	  argc -= 2;
	  
	  break;
	}
  }

  /* pass filtered argv to handle_opts() */
  newargv = PyList_New(0);
  for (i = 0; i < argc; i++) {
	PyList_Append(newargv, PyString_FromString(argv[i]));
  }
  rr = PyObject_CallMethod((PyObject *)self, \
						   "handle_opts", "O", newargv);
  for (i = 0; i < argc; i++) {
	PyObject *t = PyList_GetItem(newargv, i);
	Py_XDECREF(t);
  }
  Py_XDECREF(newargv);
  if (rr == NULL)
	return -1;
  Py_XDECREF(rr);

  /* register signal handlers */

  memset((void *)&sa, 0, sizeof(struct sigaction));
  sa.sa_handler = __sig_handler;
  sigaction (SIGINT, &sa, 0);
  sigaction (SIGPIPE, &sa, 0);
  sigaction (SIGQUIT, &sa, 0);
  sigaction (SIGTERM, &sa, 0);


  /* connect to X server */
  if (NULL == (self->display = XOpenDisplay(display_name))) {
	PyErr_SetString(PyExc_Exception, "Unable to connect to X Server");
	return -1;
  }

  if (hack_create_window(argc, argv, self)) {
	PyErr_SetString(PyExc_Exception, "Failed to create window");
	return -1;
  }

  return 0;


}

PyMODINIT_FUNC initpyhack(void) {

  PyObject *m;

  if (PyType_Ready(&GLHackType) < 0)
	return;

  m = Py_InitModule3("pyhack",
					 pyhack_functions,
					 "Python OpenGL XScreensaver Hacks module");

  if (m == NULL)
	return;

  Py_INCREF(&GLHackType);
  PyModule_AddObject(m, "GLHack", (PyObject *)&GLHackType);

}

