/*

  pyhack.h - Python XScreensaver Hack Module and GLHack Class

  Copyright (C) 2004 Alastair Tse <acnt2@cam.ac.uk>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#ifndef __PYHACK_H
#define __PYHACK_H

#include <Python.h>
#include "structmember.h"
#include "pyhack_object.h"

#ifndef PyMODINIT_FUNC  /* declarations for DLL import/export */
#define PyMODINIT_FUNC void
#endif

/* Default GL Hack options */

#define GLHACK_OPTIONS_CASES case 'r': case 'x': case 'n': hack_handle_global_opts(self, c); break;
#define GLHACK_OPTIONS_LONG {"root", 0, 0, 'r'}, {"maxfps", 1, 0, 'x'}, {"nice", 0, 0, 'n'},
#define GLHACK_OPTIONS_SHORT "rx:n"
#define GLHACK_OPTIONS_HELP "\t--root/-r\n" "\t--maxfps/-x <arg>\n" "\t--nice/-n\n"

  /* Some Convenience Macros */

#define GLHACK_ARG_KEY_FUNC(f) static PyObject * f (GLHackObject *self, PyObject *args, PyObject *kwds)
#define GLHACK_ARG_FUNC(f) static PyObject * f (GLHackObject *self, PyObject *args)
#define GLHACK_NOARGS_FUNC(f) static PyObject * f (GLHackObject *self)


  /* Function Prototypes */

static PyObject* GLHack_new(PyTypeObject *type, PyObject *args, 
							PyObject *kwds);
static void GLHack_dealloc(GLHackObject *self);
static int GLHack_init(GLHackObject *self, PyObject *args);

GLHACK_NOARGS_FUNC(GLHack_run);
GLHACK_NOARGS_FUNC(GLHack_draw);
GLHACK_ARG_FUNC(GLHack_reshape);
GLHACK_ARG_FUNC(GLHack_handle_opts);

static PyMethodDef GLHack_methods[] = {
  {"run", (PyCFunction)GLHack_run, METH_NOARGS, \
   "Run GLHack. This function returns only when the screensaver ends."},
  {"draw", (PyCFunction)GLHack_draw, METH_NOARGS, \
   "Draw a single GL Frame. You should override this function!"},
  {"reshape", (PyCFunction)GLHack_reshape, METH_VARARGS, \
   "Resize aspect ratio. You should override this function!"},
  {"handle_opts", (PyCFunction)GLHack_handle_opts, METH_VARARGS, \
   "Default Command Line Option Handler. You may override this function."},
  {NULL}
};

static PyMemberDef GLHack_members[] = {
  {"windowWidth", T_INT, offsetof(GLHackObject, windowWidth), 0, "Width of Display Window"},
  {"windowHeight", T_INT, offsetof(GLHackObject, windowHeight), 0, "Height of Display Window"},  
  {NULL}
};

static PyTypeObject GLHackType = {
  PyObject_HEAD_INIT(NULL)
  0,                             /*ob_size*/
  "pyhack.GLHack",               /*tp_name*/
  sizeof(GLHackObject),          /*tp_basicsize*/
  0,                             /*tp_itemsize*/
  (destructor)GLHack_dealloc,  /*tp_dealloc*/
  0,                             /*tp_print*/
  0,                         /*tp_getattr*/
  0,                         /*tp_setattr*/
  0,                         /*tp_compare*/
  0,                         /*tp_repr*/
  0,                         /*tp_as_number*/
  0,                         /*tp_as_sequence*/
  0,                         /*tp_as_mapping*/
  0,                         /*tp_hash */
  0,                         /*tp_call*/
  0,                         /*tp_str*/
  0,                         /*tp_getattro*/
  0,                         /*tp_setattro*/
  0,                                           /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE,    /*tp_flags*/
  "OpenGL XScreensaver Hack for Python",       /* tp_doc */
  0,                         /* tp_traverse */
  0,                         /* tp_clear */
  0,                         /* tp_richcompare */
  0,                         /* tp_weaklistoffset */
  0,                         /* tp_iter */
  0,                         /* tp_iternext */
  GLHack_methods,             /* tp_methods */
  GLHack_members,             /* tp_members */
  0,                         /* tp_getset */
  0,                         /* tp_base */
  0,                         /* tp_dict */
  0,                         /* tp_descr_get */
  0,                         /* tp_descr_set */
  0,                         /* tp_dictoffset */
  (initproc)GLHack_init,      /* tp_init */
  0,                         /* tp_alloc */
  GLHack_new,                 /* tp_new */
};

static PyMethodDef pyhack_functions[] = {
  {NULL}
};

#endif
